#ifndef _SAFEG_MEASURES_H_
#define _SAFEG_MEASURES_H_

/*
 * Comment out to enable measures
 */
// #define SAFEG_MEASUREMENTS_ENABLE

#ifdef SAFEG_MEASUREMENTS_ENABLE

/*
 * Flags to enable different methods of measurement:
 *   - GPIO: precise but requires external logic analyzer equipment
 *   - PERF: uses ARM performance monitor (CP15) to measure cycles.
 *     Then cycles are captured in RealView and multiplied by the
 *     processor clock period (ie 1/210MHz in pb1176)
 */
#define SAFEG_MEASURE_GPIO_ENA   0
#define SAFEG_MEASURE_PERF_ENA   1

/*
 * Flags to enable disable measures
 */
#define SAFEG_MEASURE_IRQ_OVH      0  /* IRQ overhead inside SafeG */
#define SAFEG_MEASURE_FIQ_OVH      0  /* TODO: FIQ overhead inside SafeG */
#define SAFEG_MEASURE_FIQ_FIQ      0  /* T-FIQ -> T-FIQ */
#define SAFEG_MEASURE_FIQ_IRQ      0  /* T-FIQ -> NT-IRQ */
#define ASP_MEASURE_IRQ_LAT        0  /* ASP interrupt latency */
#define ASP_MEASURE_TIMER_LAT      0  /* ASP TIMER interrupt latency */
#define ASP_MEASURE_ITASK_LAT      0  /* ITASK interrupt lat */
#define LINUX_MEASURE_TIMER_LAT    1  /* Linux timer handler latency */

// TODO: add more!! move ASP measure flags to ASP!

/* Maximum number of samples per measure */
#define MAX_COUNT 10000

/* GPIO constants */
#if SAFEG_MEASURE_GPIO_ENA == 1

#define GPIO_0_BASE  0x1010A000
#define GPIO_1_BASE  0x10014000

#define GPIO_DATA    0x0
#define GPIO_DIR   0x400
#define GPIO_IS    0x404
#define GPIO_IBE   0x408
#define GPIO_IEV   0x40C
#define GPIO_IE    0x410
#define GPIO_RIS   0x414
#define GPIO_MIS   0x418
#define GPIO_IC    0x41C
#define GPIO_AFSEL 0x420

#define GPIO_0_DATA_MASK_0  (GPIO_0_BASE + 4)     // 0000_0001_00
#define GPIO_0_DATA_MASK_1  (GPIO_0_BASE + 8)     // 0000_0010_00
#define GPIO_0_DATA_MASK_2  (GPIO_0_BASE + 16)    // 0000_0100_00
#define GPIO_0_DATA_MASK_3  (GPIO_0_BASE + 32)    // 0000_1000_00
#define GPIO_0_DATA_MASK_4  (GPIO_0_BASE + 64)    // 0001_0000_00
#define GPIO_0_DATA_MASK_5  (GPIO_0_BASE + 128)   // 0010_0000_00
#define GPIO_0_DATA_MASK_6  (GPIO_0_BASE + 256)   // 0100_0000_00
#define GPIO_0_DATA_MASK_7  (GPIO_0_BASE + 512)   // 1000_0000_00

#define GPIO_1_DATA_MASK_0  (GPIO_1_BASE + 4)     // 0000_0001_00
#define GPIO_1_DATA_MASK_1  (GPIO_1_BASE + 8)     // 0000_0010_00
#define GPIO_1_DATA_MASK_2  (GPIO_1_BASE + 16)    // 0000_0100_00
#define GPIO_1_DATA_MASK_3  (GPIO_1_BASE + 32)    // 0000_1000_00
#define GPIO_1_DATA_MASK_4  (GPIO_1_BASE + 64)    // 0001_0000_00
#define GPIO_1_DATA_MASK_5  (GPIO_1_BASE + 128)   // 0010_0000_00
#define GPIO_1_DATA_MASK_6  (GPIO_1_BASE + 256)   // 0100_0000_00
#define GPIO_1_DATA_MASK_7  (GPIO_1_BASE + 512)   // 1000_0000_00

#ifndef __ASSEMBLY__

/*
 * Initialize GPIO ports
 */
static inline void gpio_init(void)
{
    unsigned long *addr;

    /* clear all GPIO registers */
    for (addr = (unsigned long *) GPIO_0_BASE;
         addr < ((unsigned long *)(GPIO_0_BASE + GPIO_AFSEL)); addr++){
        *addr = 0;
    }

    for (addr = (unsigned long *) GPIO_1_BASE;
         addr < ((unsigned long *)(GPIO_1_BASE + GPIO_AFSEL)); addr++){
        *addr = 0;
    }

    /* set the direction of all pins to output */
    *((unsigned long *)(GPIO_0_BASE + GPIO_DIR)) = 0xFF;
    *((unsigned long *)(GPIO_1_BASE + GPIO_DIR)) = 0xFF;
}

/*
 * Set a port number 'no' to '1'
 */
#define GPIO_SET(base)   *((unsigned long *)(base)) = 0xFF;
#define GPIO_CLEAR(base) *((unsigned long *)(base)) = 0x00;

/*
 * Set a port number 'no' to '0' and the others to 1
 */
static inline void gpio_set_end(unsigned char no)
{
    *((unsigned long *)(GPIO_0_BASE + GPIO_AFSEL - 4))
        = (*((unsigned long *)(GPIO_0_BASE + GPIO_DATA)) & ~(1 << no));
}

#else /* __ASSEMBLY__ */

    /* Macros to set/clear a GPIO pin from assembly */

    .macro gpio_set, base, tmp, port
        mov \tmp, #255
        ldr \base, =\port
        str \tmp, [\base]
    .endm

    .macro gpio_clear, base, tmp, port
        mov \tmp, #0
        ldr \base, =\port
        str \tmp, [\base]
    .endm

#endif /* __ASSEMBLY__ */

#endif /* SAFEG_MEASURE_GPIO_ENA */

#if SAFEG_MEASURE_PERF_ENA == 1

#ifdef __ASSEMBLY__

/*
 * ---------------------------------------------------
 * MACROS for measurements using the ARM performance
 * monitor that is in the system coprocessor 15.
 *
 * All the measures are stored in an array:
 *
 *     unsigned long int safeg_latency[MAX_COUNT];
 *
 * and can be read using the RealView Debugger.
 * (NOTE: C variables are in safeg_start.c)
 * ---------------------------------------------------
 */

/*
 * perfmon_start: macro to start a measure
 *    - tmp: register that will be manipulated
 */

    .macro perfmon_start, tmp
        mov \tmp, #0
        mcr p15, 0, \tmp, c15, c12, 0  /* disable counter (writting 0) */
        mcr p15, 0, \tmp, c15, c12, 1  /* write 0 to cycle counter    */
        mov \tmp, #1
        mcr p15, 0, \tmp, c15, c12, 0  /* enable counter (writting 1) */
    .endm

/*
 * perfmon_end: macros to finish a measure
 *     - result: register to store the cycles
 *     - tmp0: register that will be manipulated
 *     - tmp1: register that will be manipulated
 */

    .macro perfmon_end, result, tmp0, tmp1
         mrc p15, 0, \result, c15 ,c12 , 1 /* result = read cycle counter */
         ldr \tmp0, =latency_counter
         ldr \tmp0, [\tmp0]                /* tmp0 = &latency_counter */
         ldr \tmp1, =MAX_COUNT             /* tmp1 = max_count (ie 10000) */
         cmp \tmp0, \tmp1                  /* if tmp0 >= tmp1 */
         subcs \tmp0, \tmp1                /*    tmp0 = tmp0 - tmp1 */
         ldr \tmp1, =safeg_latency         /* tmp1 = &safeg_latency[0] */
         str \result, [\tmp1, \tmp0, lsl #2]  /* safeg_latency[tmp0]=result */
         add \tmp0, \tmp0, #1              /* tmp0++ */
         ldr \tmp1, =latency_counter       /* tmp1 = &latency_counter */
         str \tmp0, [\tmp1]                /* *tmp1 = tmp0 */
         /* clear the instruction cache */
         mov \tmp1, #0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
         mcr p15, 0, \tmp1, c7, c5, 0
    .endm

#endif /* __ASSEMBLY__ */

#endif /* SAFEG_MEASURE_PERF_ENA */

#else /* SAFEG_MEASUREMENTS_ENABLE */

/* disable the measurement flags */
#define SAFEG_MEASURE_GPIO_ENA   0   /* don't touch */
#define SAFEG_MEASURE_PERF_ENA   0   /* don't touch */

#endif /* SAFEG_MEASUREMENTS_ENABLE */
#endif /* _SAFEG_MEASURES_H_ */
